import typer
import re
import shutil
from io import TextIOWrapper
from pathlib import Path
from gmirator import (
    DIR_ASSETS,
    DIR_SOURCE,
    DIR_HELPER,
    DIR_O_HTML,
    GMI_FOOTER
)


app = typer.Typer(help="generates your gmi and html files")

@app.command("gmi")
def gen_gmi():
    print("Generating GMI files...")
    make_index([DIR_SOURCE / "index.gmi", DIR_SOURCE / "gemlog/index.gmi"])
    make_footer(GMI_FOOTER, DIR_SOURCE)
    print("Done!")

def make_index(listing: list[Path]):
    print("Populating gemlog index ...")
    title, date = "", ""
    files = (DIR_SOURCE / "gemlog").rglob(r'*.gmi')
    files = [ \
            file for file in files \
            if re.search(r'\d\d\d\d.*\.gmi', file.as_posix()) \
    ]
    articles = ""
    for file in files[::-1]:
        title, date = find_title_date(file)
        print(f"\u2713 {date} {title.strip()}")
        articles += f"=> /gemlog/{file.stem}.gmi {date} - {title}"

    for path in listing:
        with open(path, "r+") as file:
            line = file.readline(9)
            while line != "## logs\n":
                line = file.readline(9)
            file.seek(0, 1)
            if not path.match(r"gemlog/index.gmi"):
                file.write("=> /gemlog/index.gmi all gemlogs here\n")
            file.write("\n" + articles)
            file.truncate()

def make_footer(footer_path: Path, directory: Path, footer_str: str ="\n---\n"):
    with open(footer_path) as file:
        footer = file.read()

    print("Adding footers...")
    files = directory.rglob(r'*.gmi')
    for file in files:
        with open(file, "rb+") as f:
            while line:= f.readline():
                if line == "---\n".encode("utf-8"):
                    f.seek(-5, 1)
                    break
            f.write((footer_str + footer).encode("utf-8"))
            f.truncate()
        print(f"\u2713 {file}")
    print("Footers added !")

def find_title_date(filename: Path):
    title, date = "", filename.stem[:10]
    with open(filename) as file:
        for line in file:
            if line[:2] == "# ":
                title = line[2:]
    return title, date

@app.command("html")
def gen_html():
    files = DIR_SOURCE.rglob('*.gmi')

    for file in files:
        gmi2html(file.as_posix())
    print("---\nDone processing gmi files\n")

    print("\nCopying static files...\n")
    shutil.copytree(DIR_ASSETS, DIR_O_HTML, dirs_exist_ok=True)

def gmi2html(file: str):
    new_file_path = file.replace(f"{DIR_SOURCE}", f"{DIR_O_HTML}")
    new_file_path = Path(new_file_path).with_suffix('.html')
    new_file_path.parent.mkdir(parents=True, exist_ok=True)

    print(f"Processing: {new_file_path}")

    with open(file, 'r') as gemlog:
        contents, title = process_file(gemlog)
    with open(DIR_HELPER / "header-part.html", "r") as header_file:
        header_content = header_file.read()
        header_content = re.sub("PAGETITLE", f"{title}", header_content)
    with open(DIR_HELPER / "footer-part.html", "r") as footer_file:
        footer_content = footer_file.read()

    with open(new_file_path, "w") as new_file:
        new_file.write(header_content)
        new_file.write(contents)
        new_file.write(footer_content)

def process_file(ifile: TextIOWrapper):
    in_preformated = False
    in_list = False
    contents = ""
    title = ""
    for line in ifile:
        if line == "---\n":
            break
        elif line[:2] == "# ":
            title = line[2:-1]

        newline = line

        if in_preformated:
            newline, n = re.subn(r'^```.*', "</pre>", newline)
            in_preformated = False if n else True
            contents += newline.lstrip()
            continue
        else:
            newline, n = re.subn(r'^```.*', "<pre>", newline)
            in_preformated = True if n else False

        if re.match(r'^[a-zA-Z]', newline):
            newline = "<p>" + newline.strip() + "</p>\n"
        newline, in_list = process_list(newline, in_list)
        newline = re.sub(r'^=> (?P<url>[^ ]*) ?(?P<text>.*)\n', repl_url, newline)
        newline = re.sub(r'^(#+) (.*)', repl_heading, newline)
        newline = process_inline(newline)
        contents += newline
    return contents, title

def process_list(line: str, in_list: bool):
    if in_list:
        if line[0] == "*":
            line = re.sub(r'^\* (.*)', r'<li>\1</li>', line)
        else:
            in_list = False
            line = "</ul>\n\n"
    elif line[0] == "*":
        line = "<ul>\n" + re.sub(r'^\* (.*)', r'<li>\1</li>', line)
        in_list = True
    return line, in_list

def process_inline(line: str):
    repls = {
            r'\*(.*?)\*': r'<em>\1</em>',
            r'~~(.*?)~~': r'<s>\1</s>'
    }
    for pattern, repl in repls.items():
        line = re.sub(pattern, repl, line)
    return line

def repl_url(matchobj: re.Match):
    url, text = matchobj.groupdict().values()
    if not re.findall(r'=> (http|gemini)', url):
        url = re.sub(r'gmi$', r'html', url)
    return f'<a href="{url}">=> {text if text else url}</a><br>\n'

def repl_heading(matchobj: re.Match):
    x = len(matchobj.group(1))
    return f"<h{x}>{matchobj.string.strip()}</h{x}>"

if __name__ == "__main__":
    app()
